// Generated by the protocol buffer compiler.  DO NOT EDIT!
// source: function.proto

#ifndef PROTOBUF_function_2eproto__INCLUDED
#define PROTOBUF_function_2eproto__INCLUDED

#include <string>

#include <google/protobuf/stubs/common.h>

#if GOOGLE_PROTOBUF_VERSION < 3001000
#error This file was generated by a newer version of protoc which is
#error incompatible with your Protocol Buffer headers.  Please update
#error your headers.
#endif
#if 3001000 < GOOGLE_PROTOBUF_MIN_PROTOC_VERSION
#error This file was generated by an older version of protoc which is
#error incompatible with your Protocol Buffer headers.  Please
#error regenerate this file with a newer version of protoc.
#endif

#include <google/protobuf/arena.h>
#include <google/protobuf/arenastring.h>
#include <google/protobuf/generated_message_util.h>
#include <google/protobuf/metadata.h>
#include <google/protobuf/message.h>
#include <google/protobuf/repeated_field.h>
#include <google/protobuf/extension_set.h>
#include <google/protobuf/map.h>
#include <google/protobuf/map_field_inl.h>
#include <google/protobuf/unknown_field_set.h>
#include "attr_value.pb.h"
#include "op_def.pb.h"
// @@protoc_insertion_point(includes)

namespace tensorflow {

// Internal implementation detail -- do not call these.
void protobuf_AddDesc_function_2eproto();
void protobuf_InitDefaults_function_2eproto();
void protobuf_AssignDesc_function_2eproto();
void protobuf_ShutdownFile_function_2eproto();

class FunctionDef;
class FunctionDefLibrary;
class FunctionDef_Node;
class GradientDef;

// ===================================================================

class FunctionDefLibrary : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:tensorflow.FunctionDefLibrary) */ {
 public:
  FunctionDefLibrary();
  virtual ~FunctionDefLibrary();

  FunctionDefLibrary(const FunctionDefLibrary& from);

  inline FunctionDefLibrary& operator=(const FunctionDefLibrary& from) {
    CopyFrom(from);
    return *this;
  }

  inline ::google::protobuf::Arena* GetArena() const { return GetArenaNoVirtual(); }
  inline void* GetMaybeArenaPointer() const {
    return MaybeArenaPtr();
  }
  static const ::google::protobuf::Descriptor* descriptor();
  static const FunctionDefLibrary& default_instance();

  static const FunctionDefLibrary* internal_default_instance();

  void UnsafeArenaSwap(FunctionDefLibrary* other);
  void Swap(FunctionDefLibrary* other);

  // implements Message ----------------------------------------------

  inline FunctionDefLibrary* New() const { return New(NULL); }

  FunctionDefLibrary* New(::google::protobuf::Arena* arena) const;
  void CopyFrom(const ::google::protobuf::Message& from);
  void MergeFrom(const ::google::protobuf::Message& from);
  void CopyFrom(const FunctionDefLibrary& from);
  void MergeFrom(const FunctionDefLibrary& from);
  void Clear();
  bool IsInitialized() const;

  size_t ByteSizeLong() const;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input);
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* output) const;
  ::google::protobuf::uint8* SerializeWithCachedSizesToArray(::google::protobuf::uint8* output) const {
    return InternalSerializeWithCachedSizesToArray(false, output);
  }
  int GetCachedSize() const { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const;
  void InternalSwap(FunctionDefLibrary* other);
  void UnsafeMergeFrom(const FunctionDefLibrary& from);
  protected:
  explicit FunctionDefLibrary(::google::protobuf::Arena* arena);
  private:
  static void ArenaDtor(void* object);
  inline void RegisterArenaDtor(::google::protobuf::Arena* arena);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return _internal_metadata_.arena();
  }
  inline void* MaybeArenaPtr() const {
    return _internal_metadata_.raw_arena_ptr();
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // repeated .tensorflow.FunctionDef function = 1;
  int function_size() const;
  void clear_function();
  static const int kFunctionFieldNumber = 1;
  const ::tensorflow::FunctionDef& function(int index) const;
  ::tensorflow::FunctionDef* mutable_function(int index);
  ::tensorflow::FunctionDef* add_function();
  ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef >*
      mutable_function();
  const ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef >&
      function() const;

  // repeated .tensorflow.GradientDef gradient = 2;
  int gradient_size() const;
  void clear_gradient();
  static const int kGradientFieldNumber = 2;
  const ::tensorflow::GradientDef& gradient(int index) const;
  ::tensorflow::GradientDef* mutable_gradient(int index);
  ::tensorflow::GradientDef* add_gradient();
  ::google::protobuf::RepeatedPtrField< ::tensorflow::GradientDef >*
      mutable_gradient();
  const ::google::protobuf::RepeatedPtrField< ::tensorflow::GradientDef >&
      gradient() const;

  // @@protoc_insertion_point(class_scope:tensorflow.FunctionDefLibrary)
 private:

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  friend class ::google::protobuf::Arena;
  typedef void InternalArenaConstructable_;
  typedef void DestructorSkippable_;
  ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef > function_;
  ::google::protobuf::RepeatedPtrField< ::tensorflow::GradientDef > gradient_;
  mutable int _cached_size_;
  friend void  protobuf_InitDefaults_function_2eproto_impl();
  friend void  protobuf_AddDesc_function_2eproto_impl();
  friend void protobuf_AssignDesc_function_2eproto();
  friend void protobuf_ShutdownFile_function_2eproto();

  void InitAsDefaultInstance();
};
extern ::google::protobuf::internal::ExplicitlyConstructed<FunctionDefLibrary> FunctionDefLibrary_default_instance_;

// -------------------------------------------------------------------

class FunctionDef_Node : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:tensorflow.FunctionDef.Node) */ {
 public:
  FunctionDef_Node();
  virtual ~FunctionDef_Node();

  FunctionDef_Node(const FunctionDef_Node& from);

  inline FunctionDef_Node& operator=(const FunctionDef_Node& from) {
    CopyFrom(from);
    return *this;
  }

  inline ::google::protobuf::Arena* GetArena() const { return GetArenaNoVirtual(); }
  inline void* GetMaybeArenaPointer() const {
    return MaybeArenaPtr();
  }
  static const ::google::protobuf::Descriptor* descriptor();
  static const FunctionDef_Node& default_instance();

  static const FunctionDef_Node* internal_default_instance();

  void UnsafeArenaSwap(FunctionDef_Node* other);
  void Swap(FunctionDef_Node* other);

  // implements Message ----------------------------------------------

  inline FunctionDef_Node* New() const { return New(NULL); }

  FunctionDef_Node* New(::google::protobuf::Arena* arena) const;
  void CopyFrom(const ::google::protobuf::Message& from);
  void MergeFrom(const ::google::protobuf::Message& from);
  void CopyFrom(const FunctionDef_Node& from);
  void MergeFrom(const FunctionDef_Node& from);
  void Clear();
  bool IsInitialized() const;

  size_t ByteSizeLong() const;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input);
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* output) const;
  ::google::protobuf::uint8* SerializeWithCachedSizesToArray(::google::protobuf::uint8* output) const {
    return InternalSerializeWithCachedSizesToArray(false, output);
  }
  int GetCachedSize() const { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const;
  void InternalSwap(FunctionDef_Node* other);
  void UnsafeMergeFrom(const FunctionDef_Node& from);
  protected:
  explicit FunctionDef_Node(::google::protobuf::Arena* arena);
  private:
  static void ArenaDtor(void* object);
  inline void RegisterArenaDtor(::google::protobuf::Arena* arena);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return _internal_metadata_.arena();
  }
  inline void* MaybeArenaPtr() const {
    return _internal_metadata_.raw_arena_ptr();
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const;

  // nested types ----------------------------------------------------


  // accessors -------------------------------------------------------

  // repeated string ret = 1;
  int ret_size() const;
  void clear_ret();
  static const int kRetFieldNumber = 1;
  const ::std::string& ret(int index) const;
  ::std::string* mutable_ret(int index);
  void set_ret(int index, const ::std::string& value);
  void set_ret(int index, const char* value);
  void set_ret(int index, const char* value, size_t size);
  ::std::string* add_ret();
  void add_ret(const ::std::string& value);
  void add_ret(const char* value);
  void add_ret(const char* value, size_t size);
  const ::google::protobuf::RepeatedPtrField< ::std::string>& ret() const;
  ::google::protobuf::RepeatedPtrField< ::std::string>* mutable_ret();

  // optional string op = 2;
  void clear_op();
  static const int kOpFieldNumber = 2;
  const ::std::string& op() const;
  void set_op(const ::std::string& value);
  void set_op(const char* value);
  void set_op(const char* value, size_t size);
  ::std::string* mutable_op();
  ::std::string* release_op();
  void set_allocated_op(::std::string* op);
  ::std::string* unsafe_arena_release_op();
  void unsafe_arena_set_allocated_op(
      ::std::string* op);

  // repeated string arg = 3;
  int arg_size() const;
  void clear_arg();
  static const int kArgFieldNumber = 3;
  const ::std::string& arg(int index) const;
  ::std::string* mutable_arg(int index);
  void set_arg(int index, const ::std::string& value);
  void set_arg(int index, const char* value);
  void set_arg(int index, const char* value, size_t size);
  ::std::string* add_arg();
  void add_arg(const ::std::string& value);
  void add_arg(const char* value);
  void add_arg(const char* value, size_t size);
  const ::google::protobuf::RepeatedPtrField< ::std::string>& arg() const;
  ::google::protobuf::RepeatedPtrField< ::std::string>* mutable_arg();

  // repeated string dep = 4;
  int dep_size() const;
  void clear_dep();
  static const int kDepFieldNumber = 4;
  const ::std::string& dep(int index) const;
  ::std::string* mutable_dep(int index);
  void set_dep(int index, const ::std::string& value);
  void set_dep(int index, const char* value);
  void set_dep(int index, const char* value, size_t size);
  ::std::string* add_dep();
  void add_dep(const ::std::string& value);
  void add_dep(const char* value);
  void add_dep(const char* value, size_t size);
  const ::google::protobuf::RepeatedPtrField< ::std::string>& dep() const;
  ::google::protobuf::RepeatedPtrField< ::std::string>* mutable_dep();

  // map<string, .tensorflow.AttrValue> attr = 5;
  int attr_size() const;
  void clear_attr();
  static const int kAttrFieldNumber = 5;
  const ::google::protobuf::Map< ::std::string, ::tensorflow::AttrValue >&
      attr() const;
  ::google::protobuf::Map< ::std::string, ::tensorflow::AttrValue >*
      mutable_attr();

  // @@protoc_insertion_point(class_scope:tensorflow.FunctionDef.Node)
 private:

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  friend class ::google::protobuf::Arena;
  typedef void InternalArenaConstructable_;
  typedef void DestructorSkippable_;
  ::google::protobuf::RepeatedPtrField< ::std::string> ret_;
  ::google::protobuf::RepeatedPtrField< ::std::string> arg_;
  ::google::protobuf::RepeatedPtrField< ::std::string> dep_;
  typedef ::google::protobuf::internal::MapEntryLite<
      ::std::string, ::tensorflow::AttrValue,
      ::google::protobuf::internal::WireFormatLite::TYPE_STRING,
      ::google::protobuf::internal::WireFormatLite::TYPE_MESSAGE,
      0 >
      FunctionDef_Node_AttrEntry;
  ::google::protobuf::internal::MapField<
      ::std::string, ::tensorflow::AttrValue,
      ::google::protobuf::internal::WireFormatLite::TYPE_STRING,
      ::google::protobuf::internal::WireFormatLite::TYPE_MESSAGE,
      0 > attr_;
  ::google::protobuf::internal::ArenaStringPtr op_;
  mutable int _cached_size_;
  friend void  protobuf_InitDefaults_function_2eproto_impl();
  friend void  protobuf_AddDesc_function_2eproto_impl();
  friend void protobuf_AssignDesc_function_2eproto();
  friend void protobuf_ShutdownFile_function_2eproto();

  void InitAsDefaultInstance();
};
extern ::google::protobuf::internal::ExplicitlyConstructed<FunctionDef_Node> FunctionDef_Node_default_instance_;

// -------------------------------------------------------------------

class FunctionDef : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:tensorflow.FunctionDef) */ {
 public:
  FunctionDef();
  virtual ~FunctionDef();

  FunctionDef(const FunctionDef& from);

  inline FunctionDef& operator=(const FunctionDef& from) {
    CopyFrom(from);
    return *this;
  }

  inline ::google::protobuf::Arena* GetArena() const { return GetArenaNoVirtual(); }
  inline void* GetMaybeArenaPointer() const {
    return MaybeArenaPtr();
  }
  static const ::google::protobuf::Descriptor* descriptor();
  static const FunctionDef& default_instance();

  static const FunctionDef* internal_default_instance();

  void UnsafeArenaSwap(FunctionDef* other);
  void Swap(FunctionDef* other);

  // implements Message ----------------------------------------------

  inline FunctionDef* New() const { return New(NULL); }

  FunctionDef* New(::google::protobuf::Arena* arena) const;
  void CopyFrom(const ::google::protobuf::Message& from);
  void MergeFrom(const ::google::protobuf::Message& from);
  void CopyFrom(const FunctionDef& from);
  void MergeFrom(const FunctionDef& from);
  void Clear();
  bool IsInitialized() const;

  size_t ByteSizeLong() const;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input);
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* output) const;
  ::google::protobuf::uint8* SerializeWithCachedSizesToArray(::google::protobuf::uint8* output) const {
    return InternalSerializeWithCachedSizesToArray(false, output);
  }
  int GetCachedSize() const { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const;
  void InternalSwap(FunctionDef* other);
  void UnsafeMergeFrom(const FunctionDef& from);
  protected:
  explicit FunctionDef(::google::protobuf::Arena* arena);
  private:
  static void ArenaDtor(void* object);
  inline void RegisterArenaDtor(::google::protobuf::Arena* arena);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return _internal_metadata_.arena();
  }
  inline void* MaybeArenaPtr() const {
    return _internal_metadata_.raw_arena_ptr();
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const;

  // nested types ----------------------------------------------------

  typedef FunctionDef_Node Node;

  // accessors -------------------------------------------------------

  // optional .tensorflow.OpDef signature = 1;
  bool has_signature() const;
  void clear_signature();
  static const int kSignatureFieldNumber = 1;
  private:
  void _slow_mutable_signature();
  void _slow_set_allocated_signature(
      ::google::protobuf::Arena* message_arena, ::tensorflow::OpDef** signature);
  ::tensorflow::OpDef* _slow_release_signature();
  public:
  const ::tensorflow::OpDef& signature() const;
  ::tensorflow::OpDef* mutable_signature();
  ::tensorflow::OpDef* release_signature();
  void set_allocated_signature(::tensorflow::OpDef* signature);
  ::tensorflow::OpDef* unsafe_arena_release_signature();
  void unsafe_arena_set_allocated_signature(
      ::tensorflow::OpDef* signature);

  // repeated .tensorflow.FunctionDef.Node node = 2;
  int node_size() const;
  void clear_node();
  static const int kNodeFieldNumber = 2;
  const ::tensorflow::FunctionDef_Node& node(int index) const;
  ::tensorflow::FunctionDef_Node* mutable_node(int index);
  ::tensorflow::FunctionDef_Node* add_node();
  ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef_Node >*
      mutable_node();
  const ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef_Node >&
      node() const;

  // @@protoc_insertion_point(class_scope:tensorflow.FunctionDef)
 private:

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  friend class ::google::protobuf::Arena;
  typedef void InternalArenaConstructable_;
  typedef void DestructorSkippable_;
  ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef_Node > node_;
  ::tensorflow::OpDef* signature_;
  mutable int _cached_size_;
  friend void  protobuf_InitDefaults_function_2eproto_impl();
  friend void  protobuf_AddDesc_function_2eproto_impl();
  friend void protobuf_AssignDesc_function_2eproto();
  friend void protobuf_ShutdownFile_function_2eproto();

  void InitAsDefaultInstance();
};
extern ::google::protobuf::internal::ExplicitlyConstructed<FunctionDef> FunctionDef_default_instance_;

// -------------------------------------------------------------------

class GradientDef : public ::google::protobuf::Message /* @@protoc_insertion_point(class_definition:tensorflow.GradientDef) */ {
 public:
  GradientDef();
  virtual ~GradientDef();

  GradientDef(const GradientDef& from);

  inline GradientDef& operator=(const GradientDef& from) {
    CopyFrom(from);
    return *this;
  }

  inline ::google::protobuf::Arena* GetArena() const { return GetArenaNoVirtual(); }
  inline void* GetMaybeArenaPointer() const {
    return MaybeArenaPtr();
  }
  static const ::google::protobuf::Descriptor* descriptor();
  static const GradientDef& default_instance();

  static const GradientDef* internal_default_instance();

  void UnsafeArenaSwap(GradientDef* other);
  void Swap(GradientDef* other);

  // implements Message ----------------------------------------------

  inline GradientDef* New() const { return New(NULL); }

  GradientDef* New(::google::protobuf::Arena* arena) const;
  void CopyFrom(const ::google::protobuf::Message& from);
  void MergeFrom(const ::google::protobuf::Message& from);
  void CopyFrom(const GradientDef& from);
  void MergeFrom(const GradientDef& from);
  void Clear();
  bool IsInitialized() const;

  size_t ByteSizeLong() const;
  bool MergePartialFromCodedStream(
      ::google::protobuf::io::CodedInputStream* input);
  void SerializeWithCachedSizes(
      ::google::protobuf::io::CodedOutputStream* output) const;
  ::google::protobuf::uint8* InternalSerializeWithCachedSizesToArray(
      bool deterministic, ::google::protobuf::uint8* output) const;
  ::google::protobuf::uint8* SerializeWithCachedSizesToArray(::google::protobuf::uint8* output) const {
    return InternalSerializeWithCachedSizesToArray(false, output);
  }
  int GetCachedSize() const { return _cached_size_; }
  private:
  void SharedCtor();
  void SharedDtor();
  void SetCachedSize(int size) const;
  void InternalSwap(GradientDef* other);
  void UnsafeMergeFrom(const GradientDef& from);
  protected:
  explicit GradientDef(::google::protobuf::Arena* arena);
  private:
  static void ArenaDtor(void* object);
  inline void RegisterArenaDtor(::google::protobuf::Arena* arena);
  private:
  inline ::google::protobuf::Arena* GetArenaNoVirtual() const {
    return _internal_metadata_.arena();
  }
  inline void* MaybeArenaPtr() const {
    return _internal_metadata_.raw_arena_ptr();
  }
  public:

  ::google::protobuf::Metadata GetMetadata() const;

  // nested types ----------------------------------------------------

  // accessors -------------------------------------------------------

  // optional string function_name = 1;
  void clear_function_name();
  static const int kFunctionNameFieldNumber = 1;
  const ::std::string& function_name() const;
  void set_function_name(const ::std::string& value);
  void set_function_name(const char* value);
  void set_function_name(const char* value, size_t size);
  ::std::string* mutable_function_name();
  ::std::string* release_function_name();
  void set_allocated_function_name(::std::string* function_name);
  ::std::string* unsafe_arena_release_function_name();
  void unsafe_arena_set_allocated_function_name(
      ::std::string* function_name);

  // optional string gradient_func = 2;
  void clear_gradient_func();
  static const int kGradientFuncFieldNumber = 2;
  const ::std::string& gradient_func() const;
  void set_gradient_func(const ::std::string& value);
  void set_gradient_func(const char* value);
  void set_gradient_func(const char* value, size_t size);
  ::std::string* mutable_gradient_func();
  ::std::string* release_gradient_func();
  void set_allocated_gradient_func(::std::string* gradient_func);
  ::std::string* unsafe_arena_release_gradient_func();
  void unsafe_arena_set_allocated_gradient_func(
      ::std::string* gradient_func);

  // @@protoc_insertion_point(class_scope:tensorflow.GradientDef)
 private:

  ::google::protobuf::internal::InternalMetadataWithArena _internal_metadata_;
  friend class ::google::protobuf::Arena;
  typedef void InternalArenaConstructable_;
  typedef void DestructorSkippable_;
  ::google::protobuf::internal::ArenaStringPtr function_name_;
  ::google::protobuf::internal::ArenaStringPtr gradient_func_;
  mutable int _cached_size_;
  friend void  protobuf_InitDefaults_function_2eproto_impl();
  friend void  protobuf_AddDesc_function_2eproto_impl();
  friend void protobuf_AssignDesc_function_2eproto();
  friend void protobuf_ShutdownFile_function_2eproto();

  void InitAsDefaultInstance();
};
extern ::google::protobuf::internal::ExplicitlyConstructed<GradientDef> GradientDef_default_instance_;

// ===================================================================


// ===================================================================

#if !PROTOBUF_INLINE_NOT_IN_HEADERS
// FunctionDefLibrary

// repeated .tensorflow.FunctionDef function = 1;
inline int FunctionDefLibrary::function_size() const {
  return function_.size();
}
inline void FunctionDefLibrary::clear_function() {
  function_.Clear();
}
inline const ::tensorflow::FunctionDef& FunctionDefLibrary::function(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDefLibrary.function)
  return function_.Get(index);
}
inline ::tensorflow::FunctionDef* FunctionDefLibrary::mutable_function(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDefLibrary.function)
  return function_.Mutable(index);
}
inline ::tensorflow::FunctionDef* FunctionDefLibrary::add_function() {
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDefLibrary.function)
  return function_.Add();
}
inline ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef >*
FunctionDefLibrary::mutable_function() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDefLibrary.function)
  return &function_;
}
inline const ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef >&
FunctionDefLibrary::function() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDefLibrary.function)
  return function_;
}

// repeated .tensorflow.GradientDef gradient = 2;
inline int FunctionDefLibrary::gradient_size() const {
  return gradient_.size();
}
inline void FunctionDefLibrary::clear_gradient() {
  gradient_.Clear();
}
inline const ::tensorflow::GradientDef& FunctionDefLibrary::gradient(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDefLibrary.gradient)
  return gradient_.Get(index);
}
inline ::tensorflow::GradientDef* FunctionDefLibrary::mutable_gradient(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDefLibrary.gradient)
  return gradient_.Mutable(index);
}
inline ::tensorflow::GradientDef* FunctionDefLibrary::add_gradient() {
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDefLibrary.gradient)
  return gradient_.Add();
}
inline ::google::protobuf::RepeatedPtrField< ::tensorflow::GradientDef >*
FunctionDefLibrary::mutable_gradient() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDefLibrary.gradient)
  return &gradient_;
}
inline const ::google::protobuf::RepeatedPtrField< ::tensorflow::GradientDef >&
FunctionDefLibrary::gradient() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDefLibrary.gradient)
  return gradient_;
}

inline const FunctionDefLibrary* FunctionDefLibrary::internal_default_instance() {
  return &FunctionDefLibrary_default_instance_.get();
}
// -------------------------------------------------------------------

// FunctionDef_Node

// repeated string ret = 1;
inline int FunctionDef_Node::ret_size() const {
  return ret_.size();
}
inline void FunctionDef_Node::clear_ret() {
  ret_.Clear();
}
inline const ::std::string& FunctionDef_Node::ret(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.Node.ret)
  return ret_.Get(index);
}
inline ::std::string* FunctionDef_Node::mutable_ret(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.Node.ret)
  return ret_.Mutable(index);
}
inline void FunctionDef_Node::set_ret(int index, const ::std::string& value) {
  // @@protoc_insertion_point(field_set:tensorflow.FunctionDef.Node.ret)
  ret_.Mutable(index)->assign(value);
}
inline void FunctionDef_Node::set_ret(int index, const char* value) {
  ret_.Mutable(index)->assign(value);
  // @@protoc_insertion_point(field_set_char:tensorflow.FunctionDef.Node.ret)
}
inline void FunctionDef_Node::set_ret(int index, const char* value, size_t size) {
  ret_.Mutable(index)->assign(
    reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_set_pointer:tensorflow.FunctionDef.Node.ret)
}
inline ::std::string* FunctionDef_Node::add_ret() {
  // @@protoc_insertion_point(field_add_mutable:tensorflow.FunctionDef.Node.ret)
  return ret_.Add();
}
inline void FunctionDef_Node::add_ret(const ::std::string& value) {
  ret_.Add()->assign(value);
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDef.Node.ret)
}
inline void FunctionDef_Node::add_ret(const char* value) {
  ret_.Add()->assign(value);
  // @@protoc_insertion_point(field_add_char:tensorflow.FunctionDef.Node.ret)
}
inline void FunctionDef_Node::add_ret(const char* value, size_t size) {
  ret_.Add()->assign(reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_add_pointer:tensorflow.FunctionDef.Node.ret)
}
inline const ::google::protobuf::RepeatedPtrField< ::std::string>&
FunctionDef_Node::ret() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDef.Node.ret)
  return ret_;
}
inline ::google::protobuf::RepeatedPtrField< ::std::string>*
FunctionDef_Node::mutable_ret() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDef.Node.ret)
  return &ret_;
}

// optional string op = 2;
inline void FunctionDef_Node::clear_op() {
  op_.ClearToEmpty(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline const ::std::string& FunctionDef_Node::op() const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.Node.op)
  return op_.Get(&::google::protobuf::internal::GetEmptyStringAlreadyInited());
}
inline void FunctionDef_Node::set_op(const ::std::string& value) {

  op_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), value, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set:tensorflow.FunctionDef.Node.op)
}
inline void FunctionDef_Node::set_op(const char* value) {

  op_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(value),
              GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_char:tensorflow.FunctionDef.Node.op)
}
inline void FunctionDef_Node::set_op(const char* value,
    size_t size) {

  op_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(
      reinterpret_cast<const char*>(value), size), GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_pointer:tensorflow.FunctionDef.Node.op)
}
inline ::std::string* FunctionDef_Node::mutable_op() {

  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.Node.op)
  return op_.Mutable(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* FunctionDef_Node::release_op() {
  // @@protoc_insertion_point(field_release:tensorflow.FunctionDef.Node.op)

  return op_.Release(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* FunctionDef_Node::unsafe_arena_release_op() {
  // @@protoc_insertion_point(field_unsafe_arena_release:tensorflow.FunctionDef.Node.op)
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);

  return op_.UnsafeArenaRelease(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      GetArenaNoVirtual());
}
inline void FunctionDef_Node::set_allocated_op(::std::string* op) {
  if (op != NULL) {

  } else {

  }
  op_.SetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), op,
      GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_allocated:tensorflow.FunctionDef.Node.op)
}
inline void FunctionDef_Node::unsafe_arena_set_allocated_op(
    ::std::string* op) {
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);
  if (op != NULL) {

  } else {

  }
  op_.UnsafeArenaSetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      op, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:tensorflow.FunctionDef.Node.op)
}

// repeated string arg = 3;
inline int FunctionDef_Node::arg_size() const {
  return arg_.size();
}
inline void FunctionDef_Node::clear_arg() {
  arg_.Clear();
}
inline const ::std::string& FunctionDef_Node::arg(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.Node.arg)
  return arg_.Get(index);
}
inline ::std::string* FunctionDef_Node::mutable_arg(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.Node.arg)
  return arg_.Mutable(index);
}
inline void FunctionDef_Node::set_arg(int index, const ::std::string& value) {
  // @@protoc_insertion_point(field_set:tensorflow.FunctionDef.Node.arg)
  arg_.Mutable(index)->assign(value);
}
inline void FunctionDef_Node::set_arg(int index, const char* value) {
  arg_.Mutable(index)->assign(value);
  // @@protoc_insertion_point(field_set_char:tensorflow.FunctionDef.Node.arg)
}
inline void FunctionDef_Node::set_arg(int index, const char* value, size_t size) {
  arg_.Mutable(index)->assign(
    reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_set_pointer:tensorflow.FunctionDef.Node.arg)
}
inline ::std::string* FunctionDef_Node::add_arg() {
  // @@protoc_insertion_point(field_add_mutable:tensorflow.FunctionDef.Node.arg)
  return arg_.Add();
}
inline void FunctionDef_Node::add_arg(const ::std::string& value) {
  arg_.Add()->assign(value);
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDef.Node.arg)
}
inline void FunctionDef_Node::add_arg(const char* value) {
  arg_.Add()->assign(value);
  // @@protoc_insertion_point(field_add_char:tensorflow.FunctionDef.Node.arg)
}
inline void FunctionDef_Node::add_arg(const char* value, size_t size) {
  arg_.Add()->assign(reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_add_pointer:tensorflow.FunctionDef.Node.arg)
}
inline const ::google::protobuf::RepeatedPtrField< ::std::string>&
FunctionDef_Node::arg() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDef.Node.arg)
  return arg_;
}
inline ::google::protobuf::RepeatedPtrField< ::std::string>*
FunctionDef_Node::mutable_arg() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDef.Node.arg)
  return &arg_;
}

// repeated string dep = 4;
inline int FunctionDef_Node::dep_size() const {
  return dep_.size();
}
inline void FunctionDef_Node::clear_dep() {
  dep_.Clear();
}
inline const ::std::string& FunctionDef_Node::dep(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.Node.dep)
  return dep_.Get(index);
}
inline ::std::string* FunctionDef_Node::mutable_dep(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.Node.dep)
  return dep_.Mutable(index);
}
inline void FunctionDef_Node::set_dep(int index, const ::std::string& value) {
  // @@protoc_insertion_point(field_set:tensorflow.FunctionDef.Node.dep)
  dep_.Mutable(index)->assign(value);
}
inline void FunctionDef_Node::set_dep(int index, const char* value) {
  dep_.Mutable(index)->assign(value);
  // @@protoc_insertion_point(field_set_char:tensorflow.FunctionDef.Node.dep)
}
inline void FunctionDef_Node::set_dep(int index, const char* value, size_t size) {
  dep_.Mutable(index)->assign(
    reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_set_pointer:tensorflow.FunctionDef.Node.dep)
}
inline ::std::string* FunctionDef_Node::add_dep() {
  // @@protoc_insertion_point(field_add_mutable:tensorflow.FunctionDef.Node.dep)
  return dep_.Add();
}
inline void FunctionDef_Node::add_dep(const ::std::string& value) {
  dep_.Add()->assign(value);
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDef.Node.dep)
}
inline void FunctionDef_Node::add_dep(const char* value) {
  dep_.Add()->assign(value);
  // @@protoc_insertion_point(field_add_char:tensorflow.FunctionDef.Node.dep)
}
inline void FunctionDef_Node::add_dep(const char* value, size_t size) {
  dep_.Add()->assign(reinterpret_cast<const char*>(value), size);
  // @@protoc_insertion_point(field_add_pointer:tensorflow.FunctionDef.Node.dep)
}
inline const ::google::protobuf::RepeatedPtrField< ::std::string>&
FunctionDef_Node::dep() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDef.Node.dep)
  return dep_;
}
inline ::google::protobuf::RepeatedPtrField< ::std::string>*
FunctionDef_Node::mutable_dep() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDef.Node.dep)
  return &dep_;
}

// map<string, .tensorflow.AttrValue> attr = 5;
inline int FunctionDef_Node::attr_size() const {
  return attr_.size();
}
inline void FunctionDef_Node::clear_attr() {
  attr_.Clear();
}
inline const ::google::protobuf::Map< ::std::string, ::tensorflow::AttrValue >&
FunctionDef_Node::attr() const {
  // @@protoc_insertion_point(field_map:tensorflow.FunctionDef.Node.attr)
  return attr_.GetMap();
}
inline ::google::protobuf::Map< ::std::string, ::tensorflow::AttrValue >*
FunctionDef_Node::mutable_attr() {
  // @@protoc_insertion_point(field_mutable_map:tensorflow.FunctionDef.Node.attr)
  return attr_.MutableMap();
}

inline const FunctionDef_Node* FunctionDef_Node::internal_default_instance() {
  return &FunctionDef_Node_default_instance_.get();
}
// -------------------------------------------------------------------

// FunctionDef

// optional .tensorflow.OpDef signature = 1;
inline bool FunctionDef::has_signature() const {
  return this != internal_default_instance() && signature_ != NULL;
}
inline void FunctionDef::clear_signature() {
  if (GetArenaNoVirtual() == NULL && signature_ != NULL) delete signature_;
  signature_ = NULL;
}
inline const ::tensorflow::OpDef& FunctionDef::signature() const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.signature)
  return signature_ != NULL ? *signature_
                         : *::tensorflow::OpDef::internal_default_instance();
}
inline ::tensorflow::OpDef* FunctionDef::mutable_signature() {

  if (signature_ == NULL) {
    _slow_mutable_signature();
  }
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.signature)
  return signature_;
}
inline ::tensorflow::OpDef* FunctionDef::release_signature() {
  // @@protoc_insertion_point(field_release:tensorflow.FunctionDef.signature)

  if (GetArenaNoVirtual() != NULL) {
    return _slow_release_signature();
  } else {
    ::tensorflow::OpDef* temp = signature_;
    signature_ = NULL;
    return temp;
  }
}
inline  void FunctionDef::set_allocated_signature(::tensorflow::OpDef* signature) {
  ::google::protobuf::Arena* message_arena = GetArenaNoVirtual();
  if (message_arena == NULL) {
    delete signature_;
  }
  if (signature != NULL) {
    _slow_set_allocated_signature(message_arena, &signature);
  }
  signature_ = signature;
  if (signature) {

  } else {

  }
  // @@protoc_insertion_point(field_set_allocated:tensorflow.FunctionDef.signature)
}

// repeated .tensorflow.FunctionDef.Node node = 2;
inline int FunctionDef::node_size() const {
  return node_.size();
}
inline void FunctionDef::clear_node() {
  node_.Clear();
}
inline const ::tensorflow::FunctionDef_Node& FunctionDef::node(int index) const {
  // @@protoc_insertion_point(field_get:tensorflow.FunctionDef.node)
  return node_.Get(index);
}
inline ::tensorflow::FunctionDef_Node* FunctionDef::mutable_node(int index) {
  // @@protoc_insertion_point(field_mutable:tensorflow.FunctionDef.node)
  return node_.Mutable(index);
}
inline ::tensorflow::FunctionDef_Node* FunctionDef::add_node() {
  // @@protoc_insertion_point(field_add:tensorflow.FunctionDef.node)
  return node_.Add();
}
inline ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef_Node >*
FunctionDef::mutable_node() {
  // @@protoc_insertion_point(field_mutable_list:tensorflow.FunctionDef.node)
  return &node_;
}
inline const ::google::protobuf::RepeatedPtrField< ::tensorflow::FunctionDef_Node >&
FunctionDef::node() const {
  // @@protoc_insertion_point(field_list:tensorflow.FunctionDef.node)
  return node_;
}

inline const FunctionDef* FunctionDef::internal_default_instance() {
  return &FunctionDef_default_instance_.get();
}
// -------------------------------------------------------------------

// GradientDef

// optional string function_name = 1;
inline void GradientDef::clear_function_name() {
  function_name_.ClearToEmpty(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline const ::std::string& GradientDef::function_name() const {
  // @@protoc_insertion_point(field_get:tensorflow.GradientDef.function_name)
  return function_name_.Get(&::google::protobuf::internal::GetEmptyStringAlreadyInited());
}
inline void GradientDef::set_function_name(const ::std::string& value) {

  function_name_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), value, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set:tensorflow.GradientDef.function_name)
}
inline void GradientDef::set_function_name(const char* value) {

  function_name_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(value),
              GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_char:tensorflow.GradientDef.function_name)
}
inline void GradientDef::set_function_name(const char* value,
    size_t size) {

  function_name_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(
      reinterpret_cast<const char*>(value), size), GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_pointer:tensorflow.GradientDef.function_name)
}
inline ::std::string* GradientDef::mutable_function_name() {

  // @@protoc_insertion_point(field_mutable:tensorflow.GradientDef.function_name)
  return function_name_.Mutable(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* GradientDef::release_function_name() {
  // @@protoc_insertion_point(field_release:tensorflow.GradientDef.function_name)

  return function_name_.Release(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* GradientDef::unsafe_arena_release_function_name() {
  // @@protoc_insertion_point(field_unsafe_arena_release:tensorflow.GradientDef.function_name)
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);

  return function_name_.UnsafeArenaRelease(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      GetArenaNoVirtual());
}
inline void GradientDef::set_allocated_function_name(::std::string* function_name) {
  if (function_name != NULL) {

  } else {

  }
  function_name_.SetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), function_name,
      GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_allocated:tensorflow.GradientDef.function_name)
}
inline void GradientDef::unsafe_arena_set_allocated_function_name(
    ::std::string* function_name) {
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);
  if (function_name != NULL) {

  } else {

  }
  function_name_.UnsafeArenaSetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      function_name, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:tensorflow.GradientDef.function_name)
}

// optional string gradient_func = 2;
inline void GradientDef::clear_gradient_func() {
  gradient_func_.ClearToEmpty(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline const ::std::string& GradientDef::gradient_func() const {
  // @@protoc_insertion_point(field_get:tensorflow.GradientDef.gradient_func)
  return gradient_func_.Get(&::google::protobuf::internal::GetEmptyStringAlreadyInited());
}
inline void GradientDef::set_gradient_func(const ::std::string& value) {

  gradient_func_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), value, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set:tensorflow.GradientDef.gradient_func)
}
inline void GradientDef::set_gradient_func(const char* value) {

  gradient_func_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(value),
              GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_char:tensorflow.GradientDef.gradient_func)
}
inline void GradientDef::set_gradient_func(const char* value,
    size_t size) {

  gradient_func_.Set(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), ::std::string(
      reinterpret_cast<const char*>(value), size), GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_pointer:tensorflow.GradientDef.gradient_func)
}
inline ::std::string* GradientDef::mutable_gradient_func() {

  // @@protoc_insertion_point(field_mutable:tensorflow.GradientDef.gradient_func)
  return gradient_func_.Mutable(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* GradientDef::release_gradient_func() {
  // @@protoc_insertion_point(field_release:tensorflow.GradientDef.gradient_func)

  return gradient_func_.Release(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), GetArenaNoVirtual());
}
inline ::std::string* GradientDef::unsafe_arena_release_gradient_func() {
  // @@protoc_insertion_point(field_unsafe_arena_release:tensorflow.GradientDef.gradient_func)
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);

  return gradient_func_.UnsafeArenaRelease(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      GetArenaNoVirtual());
}
inline void GradientDef::set_allocated_gradient_func(::std::string* gradient_func) {
  if (gradient_func != NULL) {

  } else {

  }
  gradient_func_.SetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(), gradient_func,
      GetArenaNoVirtual());
  // @@protoc_insertion_point(field_set_allocated:tensorflow.GradientDef.gradient_func)
}
inline void GradientDef::unsafe_arena_set_allocated_gradient_func(
    ::std::string* gradient_func) {
  GOOGLE_DCHECK(GetArenaNoVirtual() != NULL);
  if (gradient_func != NULL) {

  } else {

  }
  gradient_func_.UnsafeArenaSetAllocated(&::google::protobuf::internal::GetEmptyStringAlreadyInited(),
      gradient_func, GetArenaNoVirtual());
  // @@protoc_insertion_point(field_unsafe_arena_set_allocated:tensorflow.GradientDef.gradient_func)
}

inline const GradientDef* GradientDef::internal_default_instance() {
  return &GradientDef_default_instance_.get();
}
#endif  // !PROTOBUF_INLINE_NOT_IN_HEADERS
// -------------------------------------------------------------------

// -------------------------------------------------------------------

// -------------------------------------------------------------------


// @@protoc_insertion_point(namespace_scope)

}  // namespace tensorflow

// @@protoc_insertion_point(global_scope)

#endif  // PROTOBUF_function_2eproto__INCLUDED
