import "Windows.Media.idl";

#include <sdkddkver.h>

namespace OcvTransform
{
    [version(NTDDI_WIN8)]
    runtimeclass OcvImageManipulations
    {
        interface Windows.Media.IMediaExtension;
    }
}
